from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForSequenceClassification
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_embedded_rnn import EmbeddedRnnConfig

class EmbeddedRnnForSequenceClassification(PreTrainedModel):
    config_class = EmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)        
        self.layer = torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=config.vocab_size, embedding_dim=32),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=32, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.LSTMLastHiddenState(),
            torch.nn.Linear(in_features=32, out_features=config.num_labels)
        )

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

#오토 모델에 등록
AutoModelForSequenceClassification.register(EmbeddedRnnConfig, EmbeddedRnnForSequenceClassification)

####################

from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_embedded_rnn import EmbeddedRnnConfig

class EmbeddedRnnForFixedLengthTranslation(PreTrainedModel):
    config_class = EmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)        
        self.layer = torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=config.vocab_size, embedding_dim=32),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=32, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.SelectFromArray(index=0), #lstm output, 모든 타임 스텝(토큰: 숫자, 문자, 단어)의 숨은 상태, torch.Size([8, 380, 32]) 
            torch.nn.Linear(in_features=32, out_features=config.vocab_size)
        )

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([2, 3, 7])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            #print(labels.shape) #torch.Size([2, 3])
            logits_ = logits.view(-1, logits.size(-1))
            labels_ = labels.view(-1)
            #print(logits_.shape) #torch.Size([6, 7])
            #print(labels_.shape) #torch.Size([6])
            loss = F.nll_loss(F.log_softmax(logits_), labels_) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

'''
#오토 모델에 등록
AutoModelForFixedLengthTranslation.register(EmbeddedRnnConfig, EmbeddedRnnForFixedLengthTranslation)
'''

####################

from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForSequenceClassification
from .configuration_embedded_rnn import PretrainedEmbeddedRnnConfig
import torch
from torch.nn import functional as F
from torchtext.vocab import build_vocab_from_iterator, Vectors
import transformers

class PretrainedEmbeddedRnnForSequenceClassification(PreTrainedModel):
    config_class = PretrainedEmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)  
        self.layer = torch.nn.Sequential(
            self.load_pretrained_embedding(num_embeddings=config.vocab_size, embedding_dim=300, id_to_token=config.id_to_token),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=300, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.LSTMLastHiddenState(),
            torch.nn.Linear(in_features=32, out_features=config.num_labels)
        )

    def load_pretrained_embedding(self, num_embeddings, embedding_dim, id_to_token):
        #사이즈가 작은 파일들만 가져왔습니다. 다른 모델을 써보고 싶다면, 아래 링크를 참고해서 코드를 변경해서 사용하세요.
        #https://pytorch.org/text/stable/_modules/torchtext/vocab/vectors.html#Vectors
        '''
        #glove
        pretrained_vectors = Vectors(name='glove.6B.300d.txt', url='http://nlp.stanford.edu/data/glove.6B.zip')
        pretrained_emb = pretrained_vectors.get_vecs_by_tokens(id_to_token, lower_case_backup=True)
        embedding = torch.nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_dim).from_pretrained(pretrained_emb, freeze=False)
        return embedding
        '''
        #'''
        #fasttext
        pretrained_vectors = Vectors(name='wiki.simple.vec', url='https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.simple.vec')
        pretrained_emb = pretrained_vectors.get_vecs_by_tokens(id_to_token, lower_case_backup=True)
        embedding = torch.nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_dim).from_pretrained(pretrained_emb, freeze = False)
        return embedding
        #'''

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

#오토 모델에 등록
AutoModelForSequenceClassification.register(PretrainedEmbeddedRnnConfig, PretrainedEmbeddedRnnForSequenceClassification)

####################



from transformers import Pipeline
from transformers.pipelines import PIPELINE_REGISTRY
from torch.nn import functional as F

class TabularClassificationPipeline(Pipeline):
    def _sanitize_parameters(self, **kwargs):
        preprocess_kwargs = {}
        if "transformer_path" in kwargs:
            preprocess_kwargs["transformer_path"] = kwargs["transformer_path"]
        postprocess_kwargs = {}
        if "top_k" in kwargs:
            preprocess_kwargs["top_k"] = kwargs["top_k"]
        return preprocess_kwargs, {}, postprocess_kwargs

    def preprocess(self, inputs, transformer_path=None):
        return self.feature_extractor(inputs, transformer_path=transformer_path, return_tensors=self.framework)        

    def _forward(self, model_inputs):
        return self.model(**model_inputs)

    def postprocess(self, model_outputs, top_k=None):
        logits = model_outputs['logits']
        probabilities = F.softmax(logits, dim=-1)
        results = []
        for probability in probabilities:
            label = self.model.config.id2label[probability.argmax().item()]
            probability = probabilities[0][probability.argmax()]
            results.append({'label': label, 'score': probability.item()})   
        results.sort(key=lambda x: x['score'], reverse=True)
        if top_k != None:
            results = results[:top_k]            
        return results

PIPELINE_REGISTRY.register_pipeline('tabular-classification', 
                                    #pt_model=AutoModelForTabularClassification
                                    pipeline_class=TabularClassificationPipeline)



from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForImageClassification
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_cnn import CnnConfig
    
class CnnForImageClassification(PreTrainedModel):
    config_class = CnnConfig

    def __init__(self, config):
        super().__init__(config)
        self.layer = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=config.in_channels, out_channels=32, kernel_size=5, stride=1, padding=0),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            torch.nn.Conv2d(in_channels=32, out_channels=32, kernel_size=5, stride=1, padding=0),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            torch.nn.Flatten(), #배치를 제외한 모든 차원을 평탄화
            #pytorch_supporter.layers.LazilyInitializedLinear(out_features=config.num_labels)
            torch.nn.Linear(in_features=(512 ? config.in_channels==1 else 89888), out_features=config.num_labels)  
        )

    def forward(self, pixel_values, labels=None):
        #print(pixel_values.shape) #
        logits = self.layer(pixel_values)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            #print(labels.shape) #
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

#오토 모델에 등록
AutoModelForImageClassification.register(CnnConfig, CnnForImageClassification)

####################

from transformers import PreTrainedModel
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_cnn import CnnConfig

class CnnForKeyPointDetection(PreTrainedModel):
    config_class = CnnConfig

    def __init__(self, config):
        super().__init__(config)
        self.layer = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=3, out_channels=32, kernel_size=5, stride=1, padding=0),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            torch.nn.Conv2d(in_channels=32, out_channels=32, kernel_size=5, stride=1, padding=0),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            torch.nn.Flatten(), #배치를 제외한 모든 차원을 평탄화
            torch.nn.Linear(in_features=14112, out_features=config.num_labels * 2)
            #pytorch_supporter.layers.LazilyInitializedLinear(out_features=30)
        )

    def forward(self, pixel_values, labels=None):
        #print(pixel_values.shape) #
        logits = self.layer(pixel_values)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            #print(labels.shape) #
            #loss = torch.nn.MSELoss()(logits, labels) 
            loss = F.mse_loss(logits, labels) 
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

'''
#오토 모델에 등록
AutoModelForKeyPointDetection.register(CnnConfig, CnnForKeyPointDetection)
'''

####################



from transformers import ImageProcessingMixin
from PIL import Image
import torch
from transformers.image_utils import is_batched
from transformers import BatchFeature
from transformers import AutoImageProcessor
from torchvision import transforms
from torch.nn import functional as F

class GrayscaleImageProcessor(ImageProcessingMixin):
    def __init__(self, image_mean=[0.5], image_std=[0.5], size={"height": 28, "width": 28}, **kwargs):
        super().__init__(**kwargs)
        self.image_mean = image_mean
        self.image_std = image_std
        self.size = size 

    def __call__(self, images, return_tensors=None, **kwargs):
        composed_transforms = transforms.Compose([
            transforms.Resize((self.size['height'], self.size['width'])),
            transforms.ToTensor(), #최소 최대 정규화 ([0, 1])
            transforms.Normalize(mean=self.image_mean, std=self.image_std) #표준 정규화 ([-1, 1])
        ])
        if not is_batched(images):
            images = [images]
        images = [image.convert('L') for image in images]
        images = [composed_transforms(image) for image in images]
        #print(images[0].shape) #torch.Size([1, 28, 28])
        '''
        if return_tensors == 'pt':
            images = torch.stack(images)
        return {"pixel_values": images}
        '''
        #'''
        images = torch.stack(images)
        return BatchFeature(data={"pixel_values": images}, tensor_type=return_tensors)
        #'''

    def postprocess(self, outputs, top_k=None):
        logits = outputs['logits']
        probabilities = F.softmax(logits, dim=-1)
        results = {'labels': [], 'scores': []}
        for probability in probabilities:
            label = probability.argmax()
            probability = probability[label]
            results['labels'].append(label)
            results['scores'].append(probability)        
        return results

#오토 이미지 프로세서에 등록
AutoImageProcessor.register(GrayscaleImageProcessor, GrayscaleImageProcessor)

####################################

from transformers import ImageProcessingMixin
from PIL import Image
import torch
from transformers.image_utils import is_batched
from transformers import BatchFeature
from transformers import AutoImageProcessor
from torchvision import transforms

class KeyPointImageProcessor(ImageProcessingMixin):
    def __init__(self, image_mean=[0.5, 0.5, 0.5], image_std=[0.5, 0.5, 0.5], **kwargs):
        super().__init__(**kwargs)
        self.image_mean = image_mean
        self.image_std = image_std

    def __call__(self, images, annotations=None, return_tensors=None, **kwargs):
        return_tensors = None #return_tensors 미지원
        composed_transforms = transforms.Compose([
            transforms.ToTensor(), #최소 최대 정규화 ([0, 1])
            transforms.Normalize(mean=self.image_mean, std=self.image_std) #표준 정규화 ([-1, 1])
        ])
        if not is_batched(images):
            images = [images]
        images = [image.convert('RGB') for image in images]
        images = [composed_transforms(image) for image in images]
        #print(images[0].shape) #torch.Size([3, 28, 28])
        if annotations == None:
            '''
            if return_tensors == 'pt':
                images = torch.stack(images)
            return {'pixel_values': images}
            '''
            #'''
            return BatchFeature(data={'pixel_values': images}, tensor_type=return_tensors)
            #'''
        else:
            #if not isinstance(annotations, list):
            #    annotations = [annotations]
            if isinstance(annotations, list) and len(annotations) > 0 and not isinstance(annotations[0], list):
                annotations = [annotations]
            labels = [self.get_x_y_keypoints(keypoints) for keypoints in annotations]
            '''
            if return_tensors == 'pt':
                images = torch.stack(images)
            return {'pixel_values': images, 'labels': labels}
            '''
            #'''
            return BatchFeature(data={'pixel_values': images, 'labels': labels}, tensor_type=return_tensors)
            #'''

    def get_x_y_keypoints(self, keypoints, as_tuple=False):
        new_keypoints = []
        xs = keypoints[::3]
        ys = keypoints[1::3]
        for x, y in zip(xs, ys):
            if not as_tuple:
                new_keypoints.append(x)
                new_keypoints.append(y)
            else:
                new_keypoints.append((x, y))
        return new_keypoints

#오토 이미지 프로세서에 등록
AutoImageProcessor.register(KeyPointImageProcessor, KeyPointImageProcessor)

####################################



from . import ann
from . import cnn
from . import rnn
from . import embedded_rnn
from . import custom_bert
from . import custom_wav2vec2
from . import faster_rcnn

from .ann.configuration_ann import AnnConfig
from .ann.modeling_ann import AnnForTabularRegression
from .ann.modeling_ann import AnnForTabularBinaryClassification
from .ann.modeling_ann import AnnForTabularClassification
from .ann.feature_extraction_ann import TabularFeatureExtractor

from .cnn.configuration_cnn import CnnConfig
from .cnn.modeling_cnn import CnnForImageClassification
from .cnn.modeling_cnn import CnnForKeyPointDetection
from .cnn.image_processing_cnn import GrayscaleImageProcessor
from .cnn.image_processing_cnn import KeyPointImageProcessor

from .rnn.configuration_rnn import RnnConfig
from .rnn.modeling_rnn import RnnForAudioClassification
from .rnn.modeling_rnn import RnnForTimeSeriesRegression

from .embedded_rnn.configuration_embedded_rnn import EmbeddedRnnConfig
from .embedded_rnn.modeling_embedded_rnn import EmbeddedRnnForSequenceClassification
from .embedded_rnn.modeling_embedded_rnn import EmbeddedRnnForFixedLengthTranslation
from .embedded_rnn.modeling_embedded_rnn import PretrainedEmbeddedRnnForSequenceClassification
from .embedded_rnn.tokenization_embedded_rnn import TorchtextTokenizer

from .embedded_1dcnn.configuration_embedded_1dcnn import Embedded1dcnnConfig
from .embedded_1dcnn.modeling_embedded_1dcnn import Embedded1dcnnForSequenceClassification

from .custom_bert.configuration_custom_bert import CustomBertConfig
from .custom_bert.modeling_custom_bert import CustomBertForSequenceClassification

from .custom_wav2vec2.feature_extraction_custom_wav2vec2 import CustomWav2Vec2FeatureExtractor

from .faster_rcnn.configuration_faster_rcnn import FasterRcnnConfig
from .faster_rcnn.modeling_faster_rcnn import FasterRcnnForObjectDetection
from .faster_rcnn.image_processing_faster_rcnn import FasterRcnnImageProcessor





from . import custom_image_classification_pipeline
from . import tabular_binary_classification_pipeline
from . import tabular_classification_pipeline
from . import tabular_regression_pipeline
from . import fixed_length_translation_pipeline

from .custom_image_classification_pipeline import CustomImageClassificationPipeline
from .tabular_binary_classification_pipeline import TabularBinaryClassificationPipeline
from .tabular_classification_pipeline import TabularClassificationPipeline
from .tabular_regression_pipeline import TabularRegressionPipeline
from .fixed_length_translation_pipeline import FixedLengthTranslationPipeline



import os
import setuptools

def requirements():
    with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), encoding='utf-8') as f:
        return f.read().splitlines()

setuptools.setup(
	name='transformers-supporter',
	version='0.0.13',
	description='Transformers supporter',
	long_description=open('README.md').read(),
	long_description_content_type='text/markdown',
	author='Sang Ki Kwon',
	url='https://github.com/automatethem/transformers-supporter',
	install_requires=requirements(),
	author_email='automatethem@gmail.com',
	license='MIT',
	packages=setuptools.find_packages(),
	zip_safe=False
)



from transformers import PretrainedConfig

class RnnConfig(PretrainedConfig):
    model_type = "rnn-config"

    def __init__(self, **kwargs):
        #https://github.com/huggingface/transformers/blob/98d88b23f54e5a23e741833f1e973fdf600cc2c5/src/transformers/configuration_utils.py#L323
        #Keys are always strings in JSON so convert ids to int here.       
        super().__init__(**kwargs)
